/*
 * Copyright(C) 2008-2009 Dmitriy Kumshayev. <dq@mail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.templateit;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.apache.poi.hssf.usermodel.HSSFCell;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;

/**
 * Represents a rectangular template area in the template workbook.
 * Encapsulated accompanying data such as template parameters,  
 * columns permutation information and the original template sheet object.
 *   
 * @author Dmitriy Kumshayev
 */
class Template extends Region
{
	private static final Logger logger = Logger.getLogger(Template.class);
	private final String name;
	private final HSSFSheet sheet;
	private final List<Parameter> parameters = new LinkedList<Parameter>();
	private Map<Integer,Integer> parameterIndexMap = null;
	private final Map<String,Parameter> parametersByName = new HashMap<String, Parameter>();
	private final Map<Integer,Map<Integer,Parameter>> paramsByAbsRowCol = new HashMap<Integer,Map<Integer,Parameter>>();    
	private final List<MergeRegion> mergeRegions = new LinkedList<MergeRegion>();
	private List<Integer> selectList = null;
	private List<Integer> absSelectList = null;

	public Template(String name, HSSFSheet sheet)
	{
		this.name = name;
		this.sheet = sheet;
	}
	
	public String getName()
	{
		return name;
	}

	public void createParameter(String paramName, int r, int c, String comment)
	{
		boolean isNumber = Character.isDigit(paramName.charAt(0));
		{
				Parameter param = (isNumber ? this.getParameter(Integer
						.parseInt(paramName)) : this.getParameter(paramName));

				if (param != null)
				{
					param.setColumn(c);
					param.setRow(r);
				}
				else if (isNumber)
				{
					param = new Parameter();
					param.setColumn(c);
					param.setRow(r);
					param.setIndex(Integer.parseInt(paramName));
					param.setName("#" + paramName);
					this.addParameter(param);
				}
				else
				{
					logger.warn("Missing parameter: "+paramName +"@("+r+","+c+")");
				}
				
				if( param!=null)
				{
					int absRow = param.getRow();
					int absCol = param.getColumn();
					Map<Integer,Parameter> paramsByAbsCol = paramsByAbsRowCol.get(absRow);
					if( paramsByAbsCol==null )
					{
						paramsByAbsCol = new HashMap<Integer, Parameter>();
						paramsByAbsRowCol.put(absRow, paramsByAbsCol);
					}
					paramsByAbsCol.put(absCol, param);
					
				}
		}
	}
	
	public void addParameter(Parameter parameter)
	{
		parameters.add(parameter);
		String name = parameter.getName();
		// Parameter name is optional
		if( name != null )
		{
			if( ! parametersByName.containsKey(name) )
		    {
				parametersByName.put(name, parameter);
			}
			else
			{
				logger.warn("Duplicate parameter: "+name);
			}
		}
	}
	
	public void setParameters(Parameter[] parameters)
	{
		for( Parameter p : parameters)
		{
			addParameter(p);
		}
	}

	public Parameter getParameter(String name)
	{
		return parametersByName.get(name);
	}

	public Parameter getParameter(int idx)
	{
		return idx>=1&&idx<=parameters.size()?parameters.get(idx-1):null;
	}
	
	public int getParametersNumber()
	{
		return parameters.size();
	}

	public List<MergeRegion> getMergeRegions()
	{
		return mergeRegions;
	}
	
	public void addMergeRegion(MergeRegion mreg)
	{
		mergeRegions.add(mreg);
	}

	public void setSelectList(List<Integer> selectList)
	{
		this.selectList = selectList;

		// calculate parameterIndexMap
		if( selectList!=null)
		{
			// Initialize absolute numbers of the selected columns
			this.absSelectList = new ArrayList<Integer>(selectList.size());
			for (int c = 0; c < width(); c++)
			{
				absSelectList.add(absoluteColumn(c));
			}
			
			parameterIndexMap = new HashMap<Integer, Integer>();
			int idx = 1;
			for( int i = 1; i <= getParametersNumber(); i++ )
			{
				Parameter pi = getParameter(i);
				int pCol = pi.getColumn();
				for (int j = 0; j < absSelectList.size(); j++)
				{
					int sCol = absSelectList.get(j);
					if(pCol==sCol)
					{
						parameterIndexMap.put(i, idx++);
						break;
					}
				}
			}
		}
		else
		{
			parameterIndexMap = null;
			absSelectList = null;
		}
	}
	
	public HSSFRow getRow(int row)
	{
		return sheet.getRow(absoluteRow(row));
	}
	
	public int height()
	{
		return end().row()-start().row()+1;
	}
	
	public int width()
	{
		if( selectList != null )
		{
			return selectList.size();
		}
		else
		{
			return end().column()-start().column()+1;
		}
	}
	
	/**
	 * @param r - 0-based relative row number within the template
	 * @return
	 */
	public int absoluteRow(int r)
	{
		return start().row()+r;
	}

	/**
	 * @param r - 0-based relative column number within the template
	 * @return
	 */
	public int absoluteColumn(int c)
	{
		if( selectList != null )
		{
			return start().column()+selectList.get(c);
		}
		else
		{
			return start().column()+c;
		}
	}
	
	public Parameter getParameter(int absRow, int absCol)
	{
		Map<Integer,Parameter> paramsByAbsCol = paramsByAbsRowCol.get(absRow);
		if( paramsByAbsCol!=null )
		{
			return paramsByAbsCol.get(absCol);
		}
		return null;
	}

	public int getParameterIndex(Parameter p)
	{
		int idx = p.getIndex();
		
		if( selectList != null )
		{
			idx = parameterIndexMap.get(idx);
		}
		return idx;
	}
	
	public HSSFCell getCell(int r, int c)
	{
		HSSFRow tRow = getRow(r);
		if( tRow != null )
		{
			return tRow.getCell(absoluteColumn(c));
		}
		return null;
	}

	public int absoluteFirstRow()
	{
		return absoluteRow(0);
	}
	
	public int absoluteFirstColumn()
	{
		return absoluteColumn(0);
	}

	public int absoluteLastRow()
	{
		return absoluteRow(height()-1);
	}
	
	public int absoluteLastColumn()
	{
		return absoluteColumn(width()-1);
	}
	
	/**
	 * @return true if this Template contains the reference
	 */
	@Override
	public boolean contains(Reference reference)
	{
		if( selectList==null)
		{
			return super.contains(reference);
		}
		else
		{
			return absoluteFirstRow() <= reference.row()
			&& reference.row() <= absoluteLastRow()
			&& absoluteFirstColumn() <= reference.column()
			&& reference.column() <= absoluteLastColumn()
			&& (absSelectList!=null?absSelectList.contains(reference.column()):true);
		}
	}
	
	@Override
	public String toString()
	{
		return name+","+super.toString();
	}
	
	
}
